#!/usr/bin/perl -w
#
#############################################################################
# CopyPlaylist2 v1.0
#
#
# Small script to copy files from the amarok-playlist 
# to a certain directory.
#
# maintained by (as time permits):
# Carsten Jahn - carsten_jahn@web.de
#
# original idea and implementation:  Max Kaesbauer - max.kaesbauer@gmail.com
# custum menu item, underscore replacement:  Todd Sepke - sepketod@mtl.mit.edu
# (c) 2005
#
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################
# tabsize = 4


use warnings;
use strict;

package MainWindow;

use Qt;
use Qt::isa qw(Qt::Widget);
use Qt::slots
	refreshPlaylistInfo => [],
	configSetPicked => [],
	saveCurrentConfigSet => [],
	saveNewConfigSet => [],
	deleteCurrentConfigSet => [],
	chooseDestinationDirectory => [],
	presetOriginalClicked => [],
	presetStandardClicked => [],
	presetCheapPlayerClicked => [],
	presetDataCDClicked => [],
	namingPreviewRefresh => [],
	startClicked => [],
	cancelClicked => [];

# Definition of Qt Widgets
my $playlistInfo;
my $playlistInfoRefreshBut;

my $configSetComboBox;
my $saveCurrentConfigSetBut;
my $saveNewConfigSetBut;
my $deleteCurrentConfigSetBut;

my $destinationDirectory;
my $destinationDirectoryBut;
my $namingScheme;
my $titleLength;
my $artistLength;
my $albumLength;
my $fileNumberMinLength;
my $tracknoMinLength;
my $startNumber;
my $dirNameLength;
my $dirNameRemoveText;
my $allowedCharacters;
my $replaceSpaceWithUnderscore;
my $namingPreview;
my $modeButtonGroup;
my $modeCopy;
my $modeSymlink;
my $modeHardlink;

# Configuration Sets. This hash contains keys generated with abbreviateConfigSetName
# to make them XML::Simple compatible. The value is always a hash that lists the configuration
# options and the associated values, which are read from and written to the GUI widgets.
# The real name of the configuration set is also stored in the option hash.
my %configSets;

my $configFileName = $ENV{HOME} . "/.kde/share/apps/amarok/scripts-data/copyplaylist2-config-1.xml";


sub NEW {
	shift->SUPER::NEW(@_);
	this->setName('mainWindow');

	#
	# Construct GUI elements
	# the order of appearance influences that TAB key activation order
	#
	$playlistInfo = Qt::Label( this, 'playlistInfo' );
	$playlistInfo->setText( "..." );
	$playlistInfoRefreshBut = Qt::PushButton( "Fetch &Playlist", this, 'playlistInfoRefreshBut' );
	this->connect($playlistInfoRefreshBut, SIGNAL('clicked()'), SLOT('refreshPlaylistInfo()'));
	
	$configSetComboBox = Qt::ComboBox( this, 'configSetComboBox' );
	this->connect($configSetComboBox, SIGNAL('activated(int)'), SLOT('configSetPicked()'));
	$saveCurrentConfigSetBut = Qt::PushButton( 'Sav&e', this, 'saveCurrentConfigSetBut' );
	this->connect($saveCurrentConfigSetBut, SIGNAL 'clicked()', SLOT 'saveCurrentConfigSet()');
	$saveNewConfigSetBut = Qt::PushButton( 'Save As...', this, 'saveNewConfigSetBut' );
	this->connect($saveNewConfigSetBut, SIGNAL 'clicked()', SLOT 'saveNewConfigSet()');
	$deleteCurrentConfigSetBut = Qt::PushButton( 'Delete', this, 'deleteCurrentConfigSetBut' );
	this->connect($deleteCurrentConfigSetBut, SIGNAL 'clicked()', SLOT 'deleteCurrentConfigSet()');

	$destinationDirectory = Qt::LineEdit( this, 'destinationDirectory' );
	$destinationDirectory->setText( "/" );
	$destinationDirectoryBut = Qt::PushButton( "...", this, 'destinationDirectoryButton' );
	#$destinationDirectoryBut->setIconSet( ... where do we get the "open folder" image?
	# http://doc.trolltech.com/3.0/dirview-example.html
	this->connect($destinationDirectoryBut, SIGNAL('clicked()'), SLOT('chooseDestinationDirectory()'));
	
	$namingScheme = Qt::LineEdit( this, 'namingScheme' );
	$namingScheme->setText( "%F%A%T.%X" );
	this->connect($namingScheme, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');
	
	$titleLength = Qt::LineEdit( this, 'titleLength' );
	$titleLength->setValidator( Qt::IntValidator( 1, 100, this ) );
	$titleLength->setText( "3" );
	this->connect($titleLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');

	$artistLength = Qt::LineEdit( this, 'artistLength' );
	$artistLength->setValidator( Qt::IntValidator( 1, 100, this ) );
	$artistLength->setText( "2" );
	this->connect($artistLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');

	$albumLength = Qt::LineEdit( this, 'albumLength' );
	$albumLength->setValidator( Qt::IntValidator( 1, 100, this ) );
	$albumLength->setText( "50" );
	this->connect($albumLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');

	$tracknoMinLength = Qt::LineEdit( this, 'tracknoMinLength' );
	$tracknoMinLength->setValidator( Qt::IntValidator( 1, 8, this ) );
	$tracknoMinLength->setText( "2" );
	this->connect($tracknoMinLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');

	$fileNumberMinLength = Qt::LineEdit( this, 'fileNumberMinLength' );
	$fileNumberMinLength->setValidator( Qt::IntValidator( 1, 8, this ) );
	$fileNumberMinLength->setText( "3" );
	this->connect($fileNumberMinLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');
	
	$startNumber = Qt::LineEdit( this, 'startNumber' );
	$startNumber->setValidator( Qt::IntValidator( 0, 99999999, this ) );
	$startNumber->setText( "1" );
	this->connect($startNumber, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');
	
	$dirNameLength = Qt::LineEdit( this, 'dirNameLength' );
	$dirNameLength->setValidator( Qt::IntValidator( 1, 100, this ) );
	$dirNameLength->setText( "40" );
	this->connect($dirNameLength, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');
	$dirNameRemoveText = Qt::LineEdit( this, 'dirNameRemoveText' );
	$dirNameRemoveText->setText( "" );
	this->connect($dirNameRemoveText, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');

	$allowedCharacters = Qt::LineEdit( this, 'allowedCharacters' );
	$allowedCharacters->setText( "a-zA-Z0-9_,+" );
	this->connect($allowedCharacters, SIGNAL 'textChanged(const QString &)', SLOT 'namingPreviewRefresh()');
	$replaceSpaceWithUnderscore = Qt::CheckBox( "Replace Spaces with &Underscores _", this );
	this->connect($replaceSpaceWithUnderscore, SIGNAL 'clicked()', SLOT 'namingPreviewRefresh()');

	$namingPreview = Qt::Label( this, 'namingPreview' );

	$modeButtonGroup = Qt::HButtonGroup(this, "modeButtonGroup");
	$modeButtonGroup->setFrameStyle(Qt::Frame::NoFrame());
	$modeCopy = Qt::RadioButton( "Copy Files", $modeButtonGroup, 'modeCopy' );
	$modeSymlink = Qt::RadioButton( "Create Symbolic Links", $modeButtonGroup, 'modeSymlink' );
	$modeHardlink = Qt::RadioButton( "Create Hardlinks", $modeButtonGroup, 'modeHardlink' );
	$modeButtonGroup->setButton(0); # activate first option
	
	my $startBut = Qt::PushButton( '&Start', this, 'startBut' );
	this->connect($startBut, SIGNAL 'clicked()', SLOT 'startClicked()');
	my $cancelBut = Qt::PushButton( '&Close', this, 'cancelBut' );
	this->connect($cancelBut, SIGNAL 'clicked()', SLOT 'cancelClicked()');

	#
	# arrange GUI layout
	#
	my $maingrid = Qt::GridLayout(this, 6, 2, 10);  # height, width (# cells)
	
	my $box = Qt::HBoxLayout;
	# the variable $box holds all "sublayouts" that are constructed and then added to the grid
	$box->addWidget( $playlistInfo );
	$box->addWidget( $playlistInfoRefreshBut );
	$maingrid->addWidget(Qt::Label('amaroK Playlist',this), 0,0 );
	$maingrid->addLayout($box, 0,1 );
	Qt::ToolTip::add($playlistInfoRefreshBut, "Retrieves the current amarok playlist.\n" .
		"Click to refresh the playlist memory of copyplaylist2\n" . 
		"if you have updated the playlist.");
	
	$box = Qt::HBoxLayout;
	$box->addWidget($configSetComboBox, 1); # non-default strech 1: take all available space
	$box->addWidget($saveCurrentConfigSetBut);
	$box->addWidget($saveNewConfigSetBut);
	$box->addWidget($deleteCurrentConfigSetBut);
	$maingrid->addWidget(Qt::Label($configSetComboBox,'&Configuration Set',this), 1,0 );
	$maingrid->addLayout($box, 1,1 );
	Qt::ToolTip::add($configSetComboBox,
		"Switch between different configuration sets with this combo box.\n" . 
		"You can modify and save the configuration sets, or create new sets.\n" . 
		"All sets are saved in $configFileName\n" .
		"If the file is not found, a couple of default configuration sets are generated." );
	Qt::ToolTip::add($saveCurrentConfigSetBut,"Overwrites the configuration set with\nthe settings below.");
	Qt::ToolTip::add($saveNewConfigSetBut,"Creates a new configuration set with\nthe settings below.");
	Qt::ToolTip::add($deleteCurrentConfigSetBut,"Deletes this configuration set.");

	my $configgrid = Qt::GridLayout(6, 2, 10);  # height, width (# cells)
	my $currentrow = 0;

	$box = Qt::HBoxLayout;
	$box->addWidget( $destinationDirectory );
	$box->addWidget( $destinationDirectoryBut );
	$configgrid->addWidget(Qt::Label($destinationDirectory,"&Destination Directory",this), $currentrow,0 );
	$configgrid->addLayout($box, $currentrow++,1 );
	Qt::ToolTip::add( $destinationDirectory, "Files will be copied/linked into this directory\n" .
		"or into subdirectories if your naming scheme contains a slash /");

	#$box = Qt::HBoxLayout;
	#$box->addWidget( $presetOriginalBut );
	#$box->addWidget( $presetStandardBut );
	#$box->addWidget( $presetCheapPlayerBut );
	#$box->addWidget( $presetDataCDBut );
	#$configgrid->addWidget(Qt::Label("Naming Presets",this), $currentrow,0 );
	#$configgrid->addLayout($box, $currentrow++,1 );
	
	$configgrid->addWidget(Qt::Label($namingScheme,"Na&ming Scheme",this), $currentrow,0 );
	$configgrid->addWidget($namingScheme, $currentrow++,1 );
	Qt::ToolTip::add($namingScheme,
		"Files are created in the desination directory according\n" .
		"to this Naming Scheme. See below for the possible place holders.\n" .
		"Note that you can create directory trees simply by placing\n" .
		"slashes (\"/\") in the Naming Scheme field.");
	

	#
	$box = Qt::GridLayout(6, 6, 1);
	
	$box->addWidget(Qt::Label("%T",this), 0,0 );
	$box->addWidget(Qt::Label($titleLength, "&Title",this), 0,1 );
	$box->addWidget(Qt::Label("max. Length",this), 0,2 );
	$box->addWidget($titleLength, 0,3 );

	$box->addWidget(Qt::Label("%A",this), 1,0 );
	$box->addWidget(Qt::Label($artistLength,"&Artist",this), 1,1 );
	$box->addWidget(Qt::Label("max. Length",this), 1,2 );
	$box->addWidget($artistLength, 1,3 );
	
	$box->addWidget(Qt::Label("%L",this), 2,0 );
	$box->addWidget(Qt::Label($albumLength,"A&lbum",this), 2,1 );
	$box->addWidget(Qt::Label("max. Length",this), 2,2 );
	$box->addWidget($albumLength, 2,3 );
	
	$box->addWidget(Qt::Label("%N",this), 3,0 );
	$box->addWidget(Qt::Label($tracknoMinLength,"Album Track &Number",this), 3,1 );
	$box->addWidget(Qt::Label("min. Length",this), 3,2 );
	$box->addWidget($tracknoMinLength, 3,3 );
	
	$box->addWidget(Qt::Label("%S",this), 4,0 );
	my $label = Qt::Label($fileNumberMinLength,"Se&quential File Number",this);
	$box->addWidget($label, 4,1 );
	$box->addWidget(Qt::Label("min. Length",this), 4,2 );
	$box->addWidget($fileNumberMinLength, 4,3 );
	$box->addWidget(Qt::Label("Starting Number",this), 4,4 );
	$box->addWidget($startNumber, 4,5 );
	Qt::ToolTip::add($label,"Use %S to insert a sequence number that is incremented\nfor every song on your playlist.");

	$box->addWidget(Qt::Label("%X",this), 5,0 );
	$box->addWidget(Qt::Label("File Extension",this), 5,1 );

	$box->addWidget(Qt::Label("%F",this), 6,0 );
	$box->addWidget(Qt::Label("Original File Name (Filtered)",this), 6,1 );

	$box->addWidget(Qt::Label("%f",this), 7,0 );
	$box->addWidget(Qt::Label("Original File Name",this), 7,1 );
	
	$box->addWidget(Qt::Label("%D",this), 8,0 );
	$box->addWidget(Qt::Label($dirNameLength,"O&riginal Directory Name",this), 8,1 );
	$box->addWidget(Qt::Label("max. Length",this), 8,2 );
	$box->addWidget($dirNameLength, 8,3 );
	$box->addWidget(Qt::Label("Remove Text",this), 8,4 );
	$box->addWidget($dirNameRemoveText, 8,5 );
	Qt::ToolTip::add($dirNameRemoveText,
		"You can enter a text string here which will be removed from all path names.\n" .
		"For example, enter /home/mp3/ to remove the leading path element from your\n" .
		"filenames. This feature allows to copy whole directory trees with unchanged\n" .
		"subdirectory structure." );
	
	$configgrid->addWidget(Qt::Label("Naming Components",this), $currentrow,0 );
	$configgrid->addLayout($box, $currentrow++,1 );
	
	$box = Qt::HBoxLayout;
	$box->addWidget($allowedCharacters);
	Qt::ToolTip::add($allowedCharacters,"Characters and character groups (ranges) from\n" .
		"ID3 tags (amaroK information) that are allowed in the\n" .
		"output filename. Enter regular expression here.");
	$box->addWidget($replaceSpaceWithUnderscore);
	Qt::ToolTip::add($replaceSpaceWithUnderscore,"If selected, all blanks are replaced\n" . 
		"with the underscore _ character." );
	$configgrid->addWidget(Qt::Label($allowedCharacters,"Character &Filter",this), $currentrow,0 );
	$configgrid->addLayout($box, $currentrow++,1 );

	$configgrid->addWidget(Qt::Label("Naming Preview",this), $currentrow,0 );
	$configgrid->addWidget($namingPreview, $currentrow++,1 );
	
	$configgrid->addWidget(Qt::Label($modeButtonGroup,"File Action",this), $currentrow,0 );
	$configgrid->addWidget($modeButtonGroup, $currentrow++,1 );

	# add grid with configuration set details
	#my $configgroup = Qt::GroupBox("Configuration Set Details");
	#$configgroup->setLayout( $configgrid );
	$maingrid->addMultiCellLayout($configgrid, 2,2,0,1, Qt::AlignRight() );

	# I gave up on trying to put the configuration details in a nice frame (with border)... :-(

	#my $t = Qt::GroupBox("Bla",this, "Bla");
	#$t->setFrameStyle(Qt::Frame::Box());
	#my $l = Qt::HBoxLayout($t->layout());
	#$l->addWidget(Qt::Label("File Action",$t));
	##Qt::Label("File Action",$t);
	#$maingrid->addWidget( $t, 4,0 );

	#my $f = Qt::Frame($maingrid);
	#$f->setFrameStyle(Qt::Frame::HLine());
	#$maingrid->addWidget( $f, 4,0 );

	#my $configgroup = Qt::GroupBox("Configuration Set Details");
	#->setFrameStyle(Qt::Frame::NoFrame());
	#my $l = Qt::HBoxLayout($configgroup->layout());
	#$l->addWidget(Qt::Label("File Action",this));
	#$configgroup->setLayout($l);
	#$maingrid->addWidget( $configgroup, 4,0 );

	$box = Qt::HBoxLayout;
	$box->addWidget( $startBut );
	$box->addWidget( $cancelBut );
	$maingrid->addMultiCellLayout($box, 3,3,0,1, Qt::AlignRight() );

	# read last user settings, if available
	readConfig();

	# do initial playlist loading
	refreshPlaylistInfo();
}


my $playlist; # last read playlist
my $plcount = 0; # last read playlist count
my $plsize = 0; # last read playlist size
sub refreshPlaylistInfo()
{
	# This also reads the playlist. startClicked() is based on the playlist information.

	$plsize = 0;
	$plcount = 0;
	my $dialog1 = "";
	my $dialog2 = "";
	
	# if amarok is running ->
	# refresh current.xml using a dcop call
	my $amarokrunning = `dcop amarok 2> /dev/null`;
	chomp($amarokrunning);
	if( $amarokrunning =~ s/^$/No such application: 'amarok'/ ) {
		Qt::MessageBox::critical(this, 'amarok not found', "amarok must be running -- copyplaylist2 has to fetch the current playlist.",  Qt::MessageBox::Abort(),  Qt::MessageBox::NoButton());
		return;
	}
	my $path2currentxml = `dcop amarok playlist saveCurrentPlaylist`;
	chomp($path2currentxml);
	if( ! ($path2currentxml =~ /\// ) ) {
		# old versions of amarok just save the playlist, but do not return its path
		$path2currentxml = $ENV{HOME} . "/.kde/share/apps/amarok/current.xml";
	}

	# read XML file. The ForceArray feature stores "item" elements always as an array,
	# even if there is only one track in the playlist. (Necessary to make the loop below work.)
	eval {
		$playlist = XMLin( $path2currentxml, ForceArray => ['item'] );
	};
	if( $@ ) {
		Qt::MessageBox::critical(this, 'playlist not found', "The current playlist could not be opened at location:\n" . $path2currentxml,  Qt::MessageBox::Abort(),  Qt::MessageBox::NoButton());
		return;
	}
	my $destdir =  $destinationDirectory->text;
	chomp($destdir);
	
	# check size of playlist
	foreach my $item (@{$playlist->{item}}){
		my $plitem = substr ($item->{url}, 5);
		$plitem =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack("C",hex($1))/eg;
		my $fsize = -s $plitem;
		$plsize += $fsize;
		$plcount++;
		$dialog1 = $dialog1 . $plcount 
		. " \"" . tp($fsize) . " bytes - " 
		. basename($plitem) . " " . " \" "; 
	}
	
	# check free space at destdir -- don't fail if destdir does not exist yet.
	my $destdirAvail = 0;
	eval {
		$destdirAvail = df($destdir)->{bavail}*1024
	};
	my $freespace = tp($destdirAvail);
	my $plsizef = tp($plsize);

	
	$playlistInfo->setText("$plcount file(s), $plsizef bytes. ($freespace bytes free in destination)");
	$playlistInfo->setPaletteForegroundColor( Qt::Color(0,0,0) );
	if( $plcount == 0 || $destdirAvail < $plsize ) {
		$playlistInfo->setPaletteForegroundColor( Qt::Color(190,0,0) );
	}
	namingPreviewRefresh();
}

sub chooseDestinationDirectory
{
	my $returndir = Qt::FileDialog::getExistingDirectory(
                   $destinationDirectory->text
                   # this,
                   # "get existing directory",
                   # "Choose a directory",
                   # TRUE 
		   );
	if( length($returndir) > 0 ) {
		$destinationDirectory->setText($returndir);
		# refresh playlist / free space information
		refreshPlaylistInfo();
	}
}

sub refreshConfigSetComboBox {
	$configSetComboBox->clear();
	my @configs_sorted = sort( keys(%configSets) );
	my $conf;
	foreach $conf (@configs_sorted) {
		if( $conf =~ /ConfigSet_/ ) {  # only handle config sets here, no other elements (like currentlySelectedSet) 
			my $conf_ref = $configSets{ $conf };
			# insert configSetName for every configuration set found.
			$configSetComboBox->insertItem( $$conf_ref{configSetName} );
		}
	}
}

sub abbreviateConfigSetName {
	# this is necessary for XML-friendlyness of tags...
	my $s = $_[0];
	$s =~ s/[^a-zA-Z0-9_]+//g;
	return "ConfigSet_" . $s;
}

sub configSetPicked {
	# called when a configuration set is selected from the combo box
	# copies config data into widgets

	#print "picked config set: " . $configSetComboBox->currentText() . "\n";

	my $conf_ref = $configSets{ abbreviateConfigSetName($configSetComboBox->currentText()) };
	#print $$conf_ref{titleLength} . "\n";

	if( defined( $$conf_ref{destinationDirectory} ) ) {
		$destinationDirectory->setText( $$conf_ref{destinationDirectory} );
	}
	if( defined( $$conf_ref{mode} ) ) {
		$modeButtonGroup->setButton( $$conf_ref{mode} );
	}
	if( defined( $$conf_ref{namingScheme} ) ) {
		$namingScheme->setText( $$conf_ref{namingScheme} );
	}
	if( defined( $$conf_ref{allowedCharacters} ) ) {
		$allowedCharacters->setText( $$conf_ref{allowedCharacters} );
	}
	if( defined( $$conf_ref{startNumber} ) ) {
		$startNumber->setText( $$conf_ref{startNumber} );
	}
	if( defined( $$conf_ref{titleLength} ) ) {
		$titleLength->setText( $$conf_ref{titleLength} );
	}
	if( defined( $$conf_ref{artistLength} ) ) {
		$artistLength->setText( $$conf_ref{artistLength} );
	}
	if( defined( $$conf_ref{albumLength} ) ) {
		$albumLength->setText( $$conf_ref{albumLength} );
	}
	if( defined( $$conf_ref{fileNumberMinLength} ) ) {
		$fileNumberMinLength->setText( $$conf_ref{fileNumberMinLength} );
	}
	if( defined( $$conf_ref{tracknoMinLength} ) ) {
		$tracknoMinLength->setText( $$conf_ref{tracknoMinLength} );
	}
	if( defined($$conf_ref{dirNameLength}) ) {
		$dirNameLength->setText( $$conf_ref{dirNameLength} );
	}
	if( defined($$conf_ref{replaceSpaceWithUnderscore}) ) {
		$replaceSpaceWithUnderscore->setChecked( $$conf_ref{replaceSpaceWithUnderscore} eq "Y" ? 1 : 0 );
	}
	if( defined($$conf_ref{dirNameRemoveText}) ) {
		$dirNameRemoveText->setText( $$conf_ref{dirNameRemoveText} );
	}
	namingPreviewRefresh();

	#print $configSets{$configSetComboBox->currentText()};
	#if( defined( $$conf_ref{titleLength} ) ) {
	#	$titleLength->setText( $$conf_ref{titleLength} );
	#}
}

sub saveCurrentConfigSet {
	my $conf_ref = $configSets{abbreviateConfigSetName($configSetComboBox->currentText())};
	$$conf_ref{configSetName} = $configSetComboBox->currentText();
	$$conf_ref{destinationDirectory} = $destinationDirectory->text;
	$$conf_ref{namingScheme} = $namingScheme->text;
	$$conf_ref{startNumber} = $startNumber->text;
	$$conf_ref{allowedCharacters} = $allowedCharacters->text;
	$$conf_ref{titleLength} = $titleLength->text;
	$$conf_ref{artistLength} =$artistLength->text;
	$$conf_ref{albumLength} = $albumLength->text;
	$$conf_ref{fileNumberMinLength} = $fileNumberMinLength->text;
	$$conf_ref{tracknoMinLength} = $tracknoMinLength->text;
	$$conf_ref{dirNameLength} = $dirNameLength->text;
	$$conf_ref{dirNameRemoveText} = $dirNameRemoveText->text;
	$$conf_ref{mode} = $modeButtonGroup->id( $modeButtonGroup->selected() );		# get ID of selected buttongroup button
	$$conf_ref{replaceSpaceWithUnderscore} = ($replaceSpaceWithUnderscore->isChecked ? "Y" : "N" ); 
	# automatically write to config file (a user expects this when clicking a save button...)
	writeConfig();
}

sub saveNewConfigSet {
	my $name = Qt::InputDialog::getText( "New Configuration Set", "Enter name of new configuration set:" );
	my %newConfigSet;
	$newConfigSet{configSetName} = $name;
	# create a new key with an empty hash as configuration set
	$configSets{abbreviateConfigSetName($name)} = \%newConfigSet;
	refreshConfigSetComboBox();
	# switch to new config set
	$configSetComboBox->setCurrentText($name);
	# populate the new (current) config set with settings
	saveCurrentConfigSet();
}

sub deleteCurrentConfigSet {
	if( $configSetComboBox->count() eq 1 ) {
		Qt::MessageBox::information(this, 'Delete Configuration Set', "The last configuration set cannot be deleted.", Qt::MessageBox::Abort(), Qt::MessageBox::NoButton() );
		return;
	}

	my $name = $configSetComboBox->currentText();
	if( Qt::MessageBox::warning(this, 'Delete Configuration Set', "Really delete configuration set $name?",  Qt::MessageBox::Yes(), Qt::MessageBox::Abort() ) == Qt::MessageBox::Yes() ) {
		delete $configSets{abbreviateConfigSetName($name)};
		refreshConfigSetComboBox();
		configSetPicked();   # a new config set appears in the combo box, but it must be loaded.
		writeConfig();
	}
}

sub namingPreviewRefresh {
# $playlist undefined?
	if( defined($playlist) && $playlist->{item}->[0] ) {
		$namingPreview->setText( buildFileName( $playlist->{item}->[0], $startNumber->text ) );
	} else {
		$namingPreview->setText( "No playlist loaded or playlist empty!" );
	}
}

sub cancelClicked
{
	writeConfig();
	Qt::app()->quit();
}


use XML::Simple;
use File::Copy;
use File::Basename;
use File::Path;
use Filesys::Df;

# This is used for the naming preview as well as for the actual copy mechanism
# in startClicked().
sub buildFileName
{
	my($playlistitem);
	$playlistitem = $_[0];
	my($sequentialFileNumber);
	$sequentialFileNumber = $_[1];
	
	# TrackNo changed to Track in newer versions of amarok
	my($track);
	$track="";
	if( defined( $playlistitem->{Track} ) ) {
		$track = $playlistitem->{Track};
	} elsif( defined( $playlistitem->{TrackNo} ) ) {
		$track = $playlistitem->{TrackNo};
	} 
	
	# create sprintf format string according to users input
	my $destfile_format = $namingScheme->text;
	# replace user-friendly abbrevations (%A = Artist ...) with internal argument order
	# abbrevations: see above (labels in GUI code)
	# argument order: see below sprintf statement comments
	$destfile_format =~ s/%S/%1\$s/g;
	$destfile_format =~ s/%T/%2\$s/g;
	$destfile_format =~ s/%A/%3\$s/g;
	$destfile_format =~ s/%L/%4\$s/g;
	$destfile_format =~ s/%N/%5\$s/g;
	$destfile_format =~ s/%X/%6\$s/g;
	$destfile_format =~ s/%F/%7\$s/g;
	$destfile_format =~ s/%f/%8\$s/g;
	$destfile_format =~ s/%D/%9\$s/g;
	#print $destfile_format .  "\n";
	# "%A%T" -> "%2$s%1$s"

	my $filename = substr ($playlistitem->{url}, 5);
	# convert escaped characters (e.g. %20) to real character (e.g. blank)
	$filename =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack("C",hex($1))/eg;
	my $filesuffix = substr( $filename, rindex( $filename, '.' ) + 1 );
	
	my $dirname = $playlistitem->{Directory};
	my $remtxt = $dirNameRemoveText->text;
	$dirname =~ s/\Q$remtxt//g;
	
	my $dest_filename = sprintf(
		$destfile_format,
		sprintf('%0' . $fileNumberMinLength->text .  'd', $sequentialFileNumber),			#1
		substr( filter_string($playlistitem->{Title}), 0, $titleLength->text ),				#2
		substr( filter_string($playlistitem->{Artist}), 0, $artistLength->text ),			#3
		substr( filter_string($playlistitem->{Album}), 0, $albumLength->text ),				#4
		sprintf('%0' . $tracknoMinLength->text .  'd', $track),								#5
		$filesuffix,																		#6
		filter_string(basename( $filename )),												#7
		basename( $filename ),																#8
		substr( filter_string($dirname), 0, $dirNameLength->text )							#9
		);
	return $dest_filename
}


sub startClicked
{
	my $destdir =  $destinationDirectory->text;
	chomp($destdir);

	# check free space at destdir -- don't fail if destdir does not exist yet.
	my $destdirAvail = 0;
	eval {
		$destdirAvail = df($destdir)->{bavail}*1024
	};
	my $freespace = tp($destdirAvail);
	my $plsizef = tp($plsize);

	if( $destdirAvail < $plsize ){
		if( Qt::MessageBox::warning(this, 'Not enough diskspace available', "Not enough diskspace available!\nOnly $freespace bytes free.\nPlaylist size: $plsizef bytes.",  Qt::MessageBox::Ignore(), Qt::MessageBox::Abort() ) == Qt::MessageBox::Abort() ) {
			# abort clicked
			return;
		}
	}
	
	my $progressDialog = Qt::ProgressDialog( 'Copying Files...', 'Abort', $plcount, this, 'progressDialog', 1 ); # modal dialog
	$progressDialog->setMinimumDuration(0);
	my $progress_copiedFiles = 0;
	my $ignore_all_errors = 0;	

	# copy each playlist-item to selected directory
	my $playlist_position = $startNumber->text;
	my $dest_plitem = "";
	my $dest_filename = "";
	foreach my $item2 (@{$playlist->{item}}){
		my $plitem = $item2->{url};
		# newer amarok versions use file:///home/... instead of file:/home/...
		$plitem =~ s/file:\/\//file:/g;
		$plitem = substr ($plitem, 5);
		$plitem =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack("C",hex($1))/eg;
		$dest_filename = buildFileName($item2, $playlist_position);
		# append filename to the specified destination directory
		$dest_plitem = File::Spec->catfile( $destdir, $dest_filename );
		
		$progress_copiedFiles++;
		$progressDialog->setProgress( $progress_copiedFiles );
		$progressDialog->setLabelText( "Copying file \"$plitem\"\nto \"$dest_plitem\" ..." );
		Qt::app()->processEvents();
		if( $progressDialog->wasCancelled() != 0 ) {
			return;
		}
		
		# $dest_filename might be a relative path in case user set a naming scheme like "%A/%L/%N-%T.mp3"
		if( $dest_filename =~ /\// ) {  # $dest_filename contains at least one / slash
			# try to create directory tree (in case it does not already exist - if it exists, nothing bad happens)
			# get path name
			my $dirn = dirname( $dest_plitem );
			#print "Trying to create directory $dirn\n";
			eval { mkpath($dirn) };     # File::Path function
			if ($@) { # directory creation failed for some reason - abort this procedure
				Qt::MessageBox::critical(this, 'Could not create directory', "Could not create directory \"$dirn\":\n$@",  Qt::MessageBox::Abort(),  Qt::MessageBox::NoButton());
				return;
			}
		}
		# copy / link file
		my $action;
		my $ret = 0;
		eval {
			if( $modeCopy->isChecked() ) {
				$action = "copy";
				$ret = copy($plitem, $dest_plitem);
			} elsif( $modeSymlink->isChecked() ) {
				$action = "symlink";
				$ret = symlink($plitem, $dest_plitem);
			} elsif( $modeHardlink->isChecked() ) {
				$action = "hardlink";
				$ret = link($plitem, $dest_plitem);
			}
		};
		# successful or not? display error message and exit loop.
		# sometimes copy just gives a return code to indicate failure, sometimes it
		# dies (link exists with same filename) so handle both cases
		my $error;
		if( $@ ) {
			$error = $@;
		} elsif( $ret != 1 ) {
			$error = $!;
		}
		if( defined($error) && $ignore_all_errors eq 0 ) {
			my $response = Qt::MessageBox::warning(this, 'Could not ' . $action . ' file', "Could not " . $action . " file \"$plitem\"\nto destination \"$dest_plitem\":\n$error", "Ignore Error", "Ignore All Errors", "Abort");
			if( $response eq 1 ) {
				$ignore_all_errors = 1;
			} elsif( $response eq 2 ) {
				$progressDialog->cancel();
				return;
			}
			# case "Ignore Error" does not need special handling.
		}
		# increment file number
		$playlist_position++;
	}
	
}

sub filter_string {
	my($str);
	$str = $_[0];
	if( length($allowedCharacters->text) > 0 ) {
		# filter characters in $str only if a filter is set.
		my $re = '^' . $allowedCharacters->text;
		eval { $str =~ s/[$re]+//g };
		if( $@ ) {
			return "Invalid String";
		}
		if( $replaceSpaceWithUnderscore->isChecked ) {
			# Replace ' ' with '_'
			$str =~ s/\s/_/g;
		}
	}
	return $str;
}	

sub writeConfig
{
	# store the current set, it will be opened during readConfig.
	$configSets{currentlySelectedSet} = $configSetComboBox->currentText();

	eval { mkpath( dirname($configFileName) ) };
	if( $@ ) {
		print "error writing config file:" . $@;
		return;
	}
	my $c = \%configSets;
	XMLout( $c, NoAttr => 1, SuppressEmpty => '',  OutputFile => $configFileName );
}

sub readConfig
{
	my $cfg;
	eval { $cfg = XMLin( $configFileName, SuppressEmpty => undef ) };
	if( ! $@ ) { # file is accessible
		%configSets = %$cfg;
		my $currentSetName = $configSets{currentlySelectedSet};
		# remove the currentlySelectedSet, so that it does not show up as a configuration set in the combobox
		delete $configSets{currentlySelectedSet};
		refreshConfigSetComboBox();
		$configSetComboBox->setCurrentText( $currentSetName );
		configSetPicked(); # process as event, although the user did not really select anything (we did this programmatically)
	} else {
		print "Problems with config file, but the program continues. " . $@;
		# generate default configuration sets
		$configSets{ abbreviateConfigSetName("Standard") } = {
			'configSetName' => "Standard",
			'destinationDirectory' => "/",
			'titleLength' => "40",
			'artistLength' => "40",
			'albumLength' => "40",
			'tracknoMinLength' => "2",
			'fileNumberMinLength' => "2",
			'startNumber' => "1",
			'namingScheme'=> "%A_%L_%N-%T.%X",
			'replaceSpaceWithUnderscore' => "Y",
			'allowedCharacters'=> "a-zA-Z0-9_,+ -",
			'dirNameLength' => "40", 
			'dirNameRemoveText' => "/home/mp3/",
			'mode' => "0"
		};
		$configSets{ abbreviateConfigSetName("Cheap Player (8.3 Filenames)") } = {
			'configSetName' => "Cheap Player (8.3 Filenames)",
			'destinationDirectory' => "/",
			'titleLength' => "3",
			'artistLength' => "2",
			'albumLength' => "40",
			'tracknoMinLength' => "2",
			'fileNumberMinLength' => "3",
			'startNumber' => "1",
			'namingScheme'=> "%S%A%T.%X",
			'replaceSpaceWithUnderscore' => "N",
			'allowedCharacters'=> "a-zA-Z0-9_-",
			'dirNameLength' => "40", 
			'dirNameRemoveText' => "/home/mp3/",
			'mode' => "0"
		};
		$configSets{ abbreviateConfigSetName("Original Filenames") } = {
			'configSetName' => "Original Filenames",
			'destinationDirectory' => "/",
			'titleLength' => "40",
			'artistLength' => "40",
			'albumLength' => "40",
			'tracknoMinLength' => "2",
			'fileNumberMinLength' => "3",
			'startNumber' => "1",
			'namingScheme'=> "%f",
			'replaceSpaceWithUnderscore' => "N",
			'allowedCharacters'=> "a-zA-Z0-9_,+ -",
			'dirNameLength' => "40", 
			'dirNameRemoveText' => "/home/mp3/",
			'mode' => "0"
		};
		$configSets{ abbreviateConfigSetName("Data CD with Directory Structure") } = {
			'configSetName' => "Data CD with Directory Structure",
			'destinationDirectory' => "/",
			'titleLength' => "40",
			'artistLength' => "40",
			'albumLength' => "40",
			'tracknoMinLength' => "2",
			'fileNumberMinLength' => "3",
			'startNumber' => "1",
			'namingScheme'=> "%A - %L/%N %A %T.%X",
			'replaceSpaceWithUnderscore' => "N",
			'allowedCharacters'=> "a-zA-Z0-9_,+ -",
			'dirNameLength' => "40", 
			'dirNameRemoveText' => "/home/mp3/",
			'mode' => "0"
		};
		refreshConfigSetComboBox();
		configSetPicked(); # load one config set.
	}
}

sub tp {
	my $input = shift;
	$input = reverse $input;
	$input =~ s<(\d\d\d)(?=\d)(?!\d*\.)><$1.>g;
	return scalar reverse $input;
}


package main;
use Qt;
use MainWindow;

# Startup function
sub startup
{
  system "dcop amarok script addCustomMenuItem \"Copy Playlist Files\" \"Open CopyPlaylist2 Dialog\"";          # Add menu items to the playlist
  system "dcop amarok playlist shortStatusMessage \"CopyPlaylist2 activated -- it is accessible in the playlist context menu.\"";     # Notify user in status bar
}

# A cleanup function for stopping the script
sub cleanup
{
  # print "cleanup...";
  system "dcop amarok script removeCustomMenuItem \"Copy Playlist Files\" \"Open CopyPlaylist2 Dialog\"";       # Remove menu items from playlist
  system "dcop amarok playlist shortStatusMessage \"CopyPlaylist2 de-activated.\"";  # Say goodbye
  # print "done.\n";
  exit(0); #?
}

my $debugmode = 0;
# for development purposes, it is easier to start the script with --openwindow option
# so that amarok script stops/starts are not needed.
if( defined($ARGV[0]) && $ARGV[0] eq "--openwindow" ) {
	$debugmode = 1;
} else {
	# using stdout/err might confuse amarok
	#open(OLD_ERR,">&STDERR");
	open(STDOUT,">copyplaylist2.log");
	open(STDERR,">&STDOUT"); 
}

main::startup;
$SIG{'TERM'} = 'main::cleanup';


my $a = Qt::Application(\@ARGV);
my $mainwindow = MainWindow;
$a->setMainWidget($mainwindow);

if( $debugmode eq 1 ) {
	$mainwindow->show;
	$a->exec;
	main::cleanup();
	exit;
} else {
	# window closed until menu item is selected (endless loop, terminated by amarok)
	while(<>)
	{


		# Create wiget for playlist copy
		#
		if (/Open\ CopyPlaylist2\ Dialog/ || /configure/) {
			#print "Running copyplaylist2...";
			MainWindow::refreshPlaylistInfo();
			$mainwindow->show;
			$a->exec;
			$mainwindow->hide;
			#print "done.\n";
		}
		else {
			#print "else\n";
		}
	}
}